#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package log;
use strict;
use threads::shared;
use IO::Handle;

use constant {
	DEBUG => 0,
	INFO => 1,
	WARNING => 2,	# A bad thing might happen
	ERROR => 3,	# A bad thing happened
	CRITICAL => 4,	# One module is going down
	FATAL => 5,	# One thread is going down
	PANIC => 6,	# The entire server is going down

	LOG_OPEN => 1,
	LOG_CLOSE => 2,
	LOG_WRITE => 3,
	LOG_ROTATE => 4,
};

our $path = './logs';
our @levels = ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', 'FATAL', 'PANIC');

sub wlog($$$) {
	my ($service, $level, $text) = @_;

	my $prefix;
	$prefix = "\002\00304" if($level > INFO);
	$prefix .= $levels[$level];
	net::privmsg($main::rsnick, $config::conf{diag}, "$prefix\: ($service) $text");
	write_log('diag', '<'.$main::rsnick.'>', "$prefix\: ($service) $text");
}

our @logqueue : shared;
my %log_handles;
my %file_handles;

sub log_writer() {
        LOGQUEUE: while(1) {
		my $data;

		while(!defined($data)) {
			lock @logqueue;

			unless(@logqueue) {
				last LOGQUEUE unless thread::alive();
				cond_wait(@logqueue);
				last LOGQUEUE unless thread::alive();
			}

			$data = shift @logqueue;
		}

		my ($type, $string) = split("\0", $data, 2);
		if($type == LOG_OPEN) {
			my ($handle, $filename) = split("\0", $string);
			open_logfile($handle, $filename);
		}
		elsif($type == LOG_CLOSE) {
			my ($handle) = $string;
			close_logfile($handle);
		}
		elsif($type == LOG_WRITE) {
			my ($handle, $prefix, @payload) = split("\0", $string);
			write_logfile($handle, $prefix, @payload);
		}
		elsif($type == LOG_ROTATE) {
			rotate_logs();
		}
	}

	close_all_logs();
#	$main::status = main::ST_CLOSED();
}

sub write_logfile($$@) {
	my ($handle, $prefix, @payloads) = @_;
	unless (defined($log_handles{lc $handle})) {
		net::debug_nolog("undefined log-handle $handle, aborting write()") if main::DEBUG();
		return undef;
	}
	foreach (@payloads) {
		$_ =~ s/\003[0-9]{1,2}(?:,[0-9]{1,2})?|[[:cntrl:]]//g;
	}
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = gmtime();
	my $time = sprintf("%02d:%02d:%02d", $hour, $min, $sec);
	my $payload = $time.$prefix.' '.join("\n".$time.$prefix.' ', @payloads);
	print {$log_handles{lc $handle}} "$payload\n";
}

sub open_logfile($$) {
	my ($handle, $filename) = @_;
	if (defined($log_handles{lc $handle})) {
		net::debug_nolog("duplicate log-handle $handle, aborting open()");
		return undef;
	}
	my ($year, $month, undef, $mday) = misc::get_date();
	my $filename2 = $filename.'-'.sprintf('%04d-%02d-%02d', $year, $month, $mday);
	
	open $log_handles{lc $handle}, '>>', $path.'/'.$filename2;
	$file_handles{lc $handle} = { BASENAME => $filename, FILENAME => $filename2 };
	$log_handles{lc $handle}->autoflush(1);
}

sub close_logfile($) {
	my ($handle) = @_;
	unless (defined($log_handles{lc $handle})) {
		net::debug_nolog("undefined log-handle $handle, aborting close()");
		return undef;
	}
	close $log_handles{lc $handle};
	delete($log_handles{lc $handle});
}

sub write_log($$@) {
	my ($handle, $prefix, @payload) = @_;
	return unless @payload;

=cut
	if($main::debug) {
		foreach my $x (@payload) {
			print "LOG: $handle: $x\n";
		}
	}
=cut

	#unless(thread::alive()) { ircsendreal(@_); return; }
	#my $data = pack("cZZa", LOG_OPEN, $handle, $prefix, join("\0", @payload));
	my $data =join("\0", LOG_WRITE, $handle, $prefix, join("\0", @payload));
	{
		lock @logqueue;
		push @logqueue, $data;
		cond_signal(@logqueue);
	}
}

sub open_log($$) {
	my ($handle, $filename) = @_;
	#print "IRCSEND: ready:$ready shutdown:$thread::shutdown\n\n";
#	return unless $handle and $filename;

	if($main::debug) {
		print "LOG: Opening file $filename as $handle\n";
	}

	#unless(thread::alive()) { ircsendreal(@_); return; }
	my $data = join("\0", LOG_OPEN, $handle, $filename);
	{
		lock @logqueue;
		push @logqueue, $data;
		cond_signal(@logqueue);
	}
}

sub close_log($) {
	my ($handle) = @_;
	#print "IRCSEND: ready:$ready shutdown:$thread::shutdown\n\n";
	return unless $handle;

	if($main::debug) {
		print "LOG: Closing file-handle $handle\n";
	}

	#unless(thread::alive()) { ircsendreal(@_); return; }
	my $data = join("\0", LOG_CLOSE, $handle);
	{
		lock @logqueue;
		push @logqueue, $data;
		cond_signal(@logqueue);
	}
}

sub rotate_logs() {
	foreach my $handle (keys(%file_handles)) {
		close $log_handles{$handle};
		my ($year, $month, undef, $mday) = misc::get_date();
		$file_handles{lc $handle}{FILENAME} = $file_handles{lc $handle}{BASENAME}.'-'.sprintf('%04d-%02d-%02d', $year, $month, $mday);
		open $log_handles{$handle}, '>>', $path.'/'.$file_handles{lc $handle}{FILENAME};
	}
}

sub call_rotate_logs {
	event::addtimer('', -(time() % 86400)+86400+1, __PACKAGE__, 'log::call_rotate_logs');
	lock @logqueue;
	my $data = join("\0", LOG_ROTATE);
	push @logqueue, $data;
	cond_signal(@logqueue);
}

sub close_all_logs() {
	foreach my $handle (keys(%file_handles)) {
		close $log_handles{$handle};
		$file_handles{lc $handle} = undef;
	}
}

sub log_shutdown() {
	lock @log::logqueue;
	cond_broadcast(@log::logqueue);
}

# set a timer to rotate logs on day-change
event::addtimer('', -(time() % 86400)+86400+1, __PACKAGE__, 'log::call_rotate_logs');

1;
