#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package logserv;

use strict;
no strict 'refs';
use threads;
use threads::shared;
use Storable;

my %userlist : shared = ();
my %chanlist : shared = ();

*synced = \&net::synced;

our $lsnick = 'LogServ';

sub notice($@) {
	my $user = shift;
	
	net::notice($user->{AGENT} || $lsnick, $user->{NICK}, @_);
}
		
sub chanlog {
	my ($cn, $payload) = @_;
	log::write_log("logserv:$cn", '', $payload)
		# This if allows us to be lazy
		if defined($chanlist{lc $cn});
}

sub addchan($$) {
	my ($user, $cn) = @_;
	unless(defined($chanlist{lc $cn})) {
		log::open_log("logserv:$cn", lc($cn).'.log');
		$chanlist{lc $cn} = 1;
		net::join($lsnick, $cn);
		net::setmode($lsnick, $cn, '+o', $lsnick);
		notice($user, "Channel $cn will now be logged");
		saveconf();
		return 1;
	} else {
		notice($user, "Channel $cn is already being logged");
		return 0;
	}
}

sub delchan($$) {
	my ($user, $cn) = @_;
	if(defined($chanlist{lc $cn})) {
		log::close_log("logserv:$cn");
		delete($chanlist{lc $cn});
		net::part($lsnick, $cn, "Channel has been deleted by ".$user->{NICK});
		notice($user, "Channel $cn will not be logged");
		saveconf();
		return 1;
	} else {
		notice($user, "Channel $cn is not being logged");
		return 0;
	}
}

# Handler Functions

event::addhandler('SJOIN', undef, undef, __PACKAGE__, 'logserv::ev_sjoin');
sub ev_sjoin {
	# ($server, $cn, $ts, $chmodes, $chmodeparms, \@users, \@bans, \@excepts, \@invex);
	my (undef, $cn, undef, undef, undef, $users, undef, undef, undef) = @_;
	foreach my $user (@$users) {
		ev_join($user->{NICK}, $cn);
	}
}

event::addhandler('JOIN', undef, undef, __PACKAGE__, 'logserv::ev_join');
sub ev_join {
	my ($nick, $cn) = @_;
	return if $net::agents{lc $nick}; # Ignore agent joins.
	return unless defined($userlist{lc $nick}); # Sometimes we get JOINs after a KILL or QUIT
	{
		lock %{$userlist{lc $nick}};
		$userlist{lc $nick}{CHANS}{$cn} = 1;
	}
	if(synced()) {
		if($cn eq '0') {
			foreach my $cn (keys(%{$userlist{lc $nick}{CHANS}})) {
				ev_part($nick, $cn, 'Left all channels');
			}
		} else {
			my ($ident, $vhost) = @{$userlist{lc $nick}{INFO}};
			chanlog($cn, "-!- $nick [$ident\@$vhost] has joined $cn");
		}
	}
}

event::addhandler('PART', undef, undef, __PACKAGE__, 'logserv::ev_part');
sub ev_part {
	my ($nick, $cn, $reason) = @_;
	next if $net::agents{lc $nick}; # Ignore agent parts.
	return unless defined($userlist{lc $nick}); # Sometimes we get JOINs after a KILL or QUIT
	{
		lock %{$userlist{lc $nick}};
		delete($userlist{lc $nick}{CHANS}{$cn});
	}
	my ($ident, $vhost) = @{$userlist{lc $nick}{INFO}};
	chanlog("$cn", "-!- $nick [$ident\@$vhost] has left $cn [$reason]");
}

event::addhandler('KICK', undef, undef, __PACKAGE__, 'logserv::ev_kick');
sub ev_kick {
	my ($src, $cn, $target, $reason) = @_;
	return unless defined($userlist{lc $target}); # Sometimes we get JOINs after a KILL or QUIT
	if(lc $target eq lc $lsnick) {
		net::agent_join($lsnick, $cn);
		net::setmode($lsnick, $cn, '+o', $lsnick);
		return;
	}
	{
		lock %{$userlist{lc $target}};
		delete($userlist{lc $target}{CHANS}{$cn});
	}
	chanlog("$cn", "-!- $target was kicked by $src [$reason]");
}

event::addhandler('MODE', undef, undef, __PACKAGE__, 'logserv::ev_mode');
sub ev_mode {
	my ($src, $cn, $modes, $parms) = @_;
	return unless synced();
	chanlog("$cn", "-!- mode/$cn [$modes".($parms ? " $parms" : '')."] by $src");
}

event::addhandler('NICKCONN', undef, undef, __PACKAGE__, 'logserv::ev_nickconn');
sub ev_nickconn {
	my ($nick, $ident, $host, $modes, $vhost, $cloakhost) = @_[0,3,4,7,8,11];
        if ($vhost eq '*') {
                if ({modes::splitumodes($modes)}->{x} eq '+') {
                        if(defined($cloakhost)) {
                                $vhost = $cloakhost;
                        }
                        else {
				# Since we have no desire to do net::userhost checks
				# This makes us dependent on VHP or CLK.
				# Do we care? Not at the moment.
				# This should NEVER happen with VHP or CLK.
				$vhost = $host;
                        }
                } else {
                        $vhost = $host;
                }
        }
	lock %userlist;
	$userlist{lc $nick} = &share({});
	$userlist{lc $nick}{INFO} = &share([]);
	push @{$userlist{lc $nick}{INFO}}, $ident, $vhost;
	$userlist{lc $nick}{CHANS} = &share({});
}

event::addhandler('NICKCHANGE', undef, undef, __PACKAGE__, 'logserv::ev_nickchange');
sub ev_nickchange {
	my ($old, $new) = @_;
	return unless defined($userlist{lc $old}); # Sometimes we get JOINs after a KILL or QUIT
	unless (lc($old) eq lc($new)) {
		lock %userlist;
		lock %{$userlist{lc $old}};
		$userlist{lc $new} = $userlist{lc $old};
		delete($userlist{lc $old});
	}
	foreach my $cn (keys(%{$userlist{lc $new}{CHANS}})) {
		chanlog($cn, "-!- $old is now known as $new");
	}
}

event::addhandler('QUIT', undef, undef, __PACKAGE__, 'logserv::ev_quit');
sub ev_quit {
	my ($nick, $reason) = @_;
	my ($ident, $vhost) = @{$userlist{lc $nick}{INFO}};
	if (synced()) {
		foreach my $cn (keys(%{$userlist{lc $nick}{CHANS}})) {
			chanlog($cn, "$nick [$ident\@$vhost] has quit [$reason]");
		}
	}
	lock %userlist;
	delete($userlist{lc $nick});
}

event::addhandler('PRIVMSG', undef, '#', __PACKAGE__, 'logserv::ev_message');
sub ev_message {
	my ($nick, $cn, $message) = @_;
	return unless defined($chanlist{lc $cn});
	chanlog($cn, "<$nick> $message");
}
event::addhandler('NOTICE', undef, '#', __PACKAGE__, 'logserv::ev_notice');
sub ev_notice {
	my ($nick, $cn, $message) = @_;
	return unless defined($chanlist{lc $cn});
	chanlog($cn, "-$nick:$cn- $message");
}

event::addhandler('CHGHOST', undef, undef, __PACKAGE__, 'logserv::ev_chghost');
sub ev_chghost {
	my (undef, $nick, $vhost) = @_;
	return unless defined($userlist{lc $nick}); # Sometimes we get JOINs after a KILL or QUIT
	{
		lock %{$userlist{lc $nick}};
		my ($ident, undef) = @{$userlist{lc $nick}{INFO}};
		delete $userlist{lc $nick}{INFO};
		$userlist{lc $nick}{INFO} = &share([]);
		push @{$userlist{lc $nick}{INFO}}, $ident, $vhost;
	}
	
}

event::addhandler('KILL', undef, undef, __PACKAGE__, 'logserv::ev_kill');
sub ev_kill {
	my ($src, $target, $reason) = @_;
	next if $net::agents{lc $target} or !defined($userlist{lc $target}); # Ignore agent kills.
	my ($ident, $vhost) = @{$userlist{lc $target}{INFO}};
	if (synced()) {
		foreach my $cn (keys(%{$userlist{lc $target}{CHANS}})) {
			chanlog($cn, "$target [$ident\@$vhost] has quit [Killed ($src ($reason))]");
		}
	}
	lock %userlist;
	delete($userlist{lc $target});
}

event::addhandler('TOPIC', undef, undef, __PACKAGE__, 'logserv::ev_topic');
sub ev_topic {
	my ($src, $cn, $setter, undef, $topic) = @_;
	# We don't care about the timestamp
	return unless synced();
	chanlog($cn, "$src changed the topic of $cn to: $topic".($setter ne $src ? " ($setter)" : ''));
}

event::addhandler('SEOS', undef, undef, __PACKAGE__, 'logserv::ev_connect');
sub ev_connect {
	net::agent_connect($lsnick, 'services', undef, '+pqzBHSD', 'Log Service');
	net::join($lsnick, $config::conf{diag});
	net::setmode($lsnick, $config::conf{diag}, '+o', $lsnick);
	join_chans();
}

# Internal Only functions.

sub saveconf() {
	my @channels = keys(%chanlist);
	Storable::nstore(\@channels, "config/logserv/chans.conf");
}

sub loadconf() {
	my @channels = @{Storable::retrieve("config/logserv/chans.conf")};
	foreach my $cn (@channels) {
		$chanlist{lc $cn} = 1;
	}
}

sub join_chans() {
	foreach my $cn (keys(%chanlist)) {
		log::open_log("logserv:$cn", lc($cn).'.log');
		net::join($lsnick, $cn);
		net::setmode($lsnick, $cn, '+o', $lsnick);
	}
}

sub init { loadconf(); }
sub begin { }
sub end { }
sub unload { saveconf(); }

1;
